/*
	GNUJSP - a free JSP implementation
	Copyright (C) 1998-1999, Vincent Partington <vincentp@xs4all.nl>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package	org.gjt.jsp;

import	java.io.File;
import	java.io.IOException;
import	java.io.PrintWriter;
import	java.lang.reflect.Field;
import	java.lang.reflect.InvocationTargetException;
import	java.lang.reflect.Method;
import	java.util.Enumeration;
import	java.util.Hashtable;
import	java.util.Vector;
import	javax.servlet.Servlet;
import	javax.servlet.ServletConfig;
import	javax.servlet.ServletContext;
import	javax.servlet.ServletException;
import	javax.servlet.http.HttpServlet;
import	javax.servlet.http.HttpServletRequest;
import	javax.servlet.http.HttpServletResponse;

public class JSPServlet extends HttpServlet {
	private File			classFilesDir;
	private boolean			checkUri;
	private String			jspExtension;
	private JSPClassLoader	jspClassLoader;
	private Hashtable		jspServlets;
	private String			pageBase;
	private String[]		compiler;

	public void init(ServletConfig config) throws ServletException {
		super.init(config);

		String	s;

		s = config.getInitParameter("classfiles");
		if(s == null) {
			s = config.getInitParameter("compiler.classpath");
			if(s == null) {
				throw new ServletException("no compiler.classpath init parameter specified");
			}
		}
		classFilesDir = new File(s);
		try {
			classFilesDir = new File(classFilesDir.getCanonicalPath());
		} catch(IOException ioexc) {
			classFilesDir = new File(classFilesDir.getAbsolutePath());
		}

		s = config.getInitParameter("checkuri");
		if(s != null && s.equalsIgnoreCase("true")) {
			checkUri = true;
		} else {
			checkUri = false;
		}

		s = config.getInitParameter("extension");
		if(s != null) {
			if(s.charAt(0) == '.') {
				jspExtension = s;
			} else {
				jspExtension = "." + s;
			}
		} else {
			jspExtension = ".jsp";
		}

 		s = config.getInitParameter("pagebase");
		if(s != null) {
			pageBase = s;
		} else {
			pageBase = "";
		}

		s = config.getInitParameter("compiler");

		if(s != null) {
			boolean	fileAdded	= false;
			Vector	v			= new Vector();
			v.addElement(s);

			String	classpath	= System.getProperty("java.class.path") + File.pathSeparator + classFilesDir.toString();

			s = config.getInitParameter("compiler.options");

			while(s != null && s.length() > 0) {
				int		i	= s.indexOf(" ");

				if(i == -1) {
					i = s.length();
				}

				String	s2	= s.substring(0, i).trim();
				s = s.substring(i).trim();

				if(s2.equals("%1"))
				{
					v.addElement(s2);
					fileAdded = true;
				} else if(s2.equals("%classpath%")) {
					v.addElement(classpath);
				} else if(s2.startsWith("%") && s2.endsWith("%")) {
					s2 = s2.substring(1, s2.length() - 1);
					String	s3	= config.getInitParameter("compiler." + s2);

					if(s3 != null) {
						v.addElement(s3);
					} else {
						throw new ServletException("Unknown compiler.options parameter: " + s2);
					}
				} else {
					v.addElement(s2);
				}
			}

			if(!fileAdded) {
				v.addElement("%1");
			}

			compiler = new String[v.size()];

			for(int i = 0; i < compiler.length; i++) {
				compiler[i] = (String) v.elementAt(i);
			}
		}

		jspClassLoader = new JSPClassLoader(classFilesDir);
		jspServlets = new Hashtable();
	}

	protected void service(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
		int				i, j;
		File			jspFile;
		String			s, className;
		Servlet			srv;
		JSPCompiler		jspCompiler;

		try {
			// check URI for security
			if(checkUri) {
				s = urlDecode(request.getRequestURI());
				i = s.indexOf("/servlet/");
				j = s.indexOf("/jsp/");
				if(i >= 0 && j >= 0 && i < j) {
					throw new JSPException("Direct requests to /servlet/jsp are not allowed.");
				}
			}

			// determine jsp file and class name
			s = request.getServletPath();
			if(s == null) {
				s = request.getRequestURI();
			}
			if(s != null && s.endsWith(jspExtension)) {
				jspFile = new File(getServletContext().getRealPath(s));
			} else {
				s = request.getPathTranslated();
				if(s == null) {
					throw new JSPException("No JSP page specified");
				}
				if(!s.endsWith(jspExtension)) {
					throw new JSPException("The requested file does not have the required extension " + jspExtension);
				}
				// FIXME: should urlDecode() be used here? Seems to be necessary for ServletExec, but will it work with % in the filename?
				jspFile = new File(pageBase + urlDecode(s));
			}
			try {
				jspFile = new File(jspFile.getCanonicalPath());
			} catch(IOException ioexc) {
				jspFile = new File(jspFile.getAbsolutePath());
			}
			className = JSPCompiler.getClassNameForJspFile(jspFile);

			// try to load jsp servlet from cache or class file
			srv = (Servlet) jspServlets.get(className);
			if(srv == null) {
				srv = loadServlet(className);
			}

			// unload servlet if dependencies demand it
			if(srv != null) {
				if(checkServletDependencies(srv.getClass())) {
					unloadServlet(srv);
					refreshClassLoader(srv.getClass());
					srv = null;
				}
			}

			// compile & load jsp servlet if we haven't loaded it yet
			if(srv == null) {
				jspCompiler = new JSPCompiler(classFilesDir, jspFile, className, compiler, request);
				jspCompiler.compile();

				srv = loadServlet(className);
				if(srv == null) {
					throw new JSPException("Could not load jsp servlet class " + className);
				}
			}
		} catch(JSPException jspexc) {
			response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "JSP compile-time error");
			response.setContentType("text/html");
			PrintWriter out = response.getWriter();
			out.println("<HTML><HEAD><TITLE>" + HttpServletResponse.SC_INTERNAL_SERVER_ERROR + " JSP compile-time error</TITLE></HEAD><BODY>" +
						"<H2>" + HttpServletResponse.SC_INTERNAL_SERVER_ERROR + " JSP compile-time error</H2>" +
						"The JSP page you requested could not be served because the following error(s) occured:<BR><PRE>");
			out.println(jspexc.getMessage());
			out.println("</PRE></BODY></HTML>");
			out.close();
			return;
		}

		// run jsp servlet
		srv.service(request, response);
	}

	public void destroy() {
		super.destroy();

		Enumeration		e;
		Servlet			srv;

		e = jspServlets.elements();
		while(e.hasMoreElements()) {
			srv = (Servlet) e.nextElement();
			unloadServlet(srv);
		}
	}

	public String getServletInfo() {
		return "GNUJSP JSPServlet";
	}

	private Servlet loadServlet(String className) throws JSPException {
		Class		c;
		Field		fld;
		Servlet		srv;

		// load servlet class
		try {
			c = jspClassLoader.loadClass(className);
			if(c == null)
				return null;
		} catch(ClassNotFoundException cnfexc) {
			return null;
		}

		// check compiler version dependency
		try {
			fld = c.getField("__compilerVersionNr");
			if(fld.getInt(null) != JSPCompiler.COMPILER_VERSION_NR) {
				refreshClassLoader(c);
				return null;
			}
		} catch(IllegalAccessException fiaexc) {
			throw new JSPException("Jsp servlet class " + className + " does not have a public static __compilerVersionNr field");
		} catch(NoSuchFieldException nsfexc) {
			throw new JSPException("Jsp servlet class " + className + " does not have a __compilerVersionNr field of type int");
		} catch(SecurityException secexc) {
			throw new JSPException("Jsp servlet class " + className + " has a security problem with its __compilerVersionNr field");
		}

		// check servlet's other dependencies
		if(checkServletDependencies(c)) {
			refreshClassLoader(c);
			return null;
		}

		// instantiate servlet
		try {
			srv = (Servlet) c.newInstance();
		} catch(IllegalAccessException iaexc) {
			throw new JSPException("Could not instantiate jsp servlet class " + className + " because of an illegal access exception");
		} catch(InstantiationException iexc) {
			throw new JSPException("Could not instantiate jsp servlet class " + className + " because it is an interface or an abstract class");
		}

		// init servlet
		try {
			srv.init(new ServletConfig() {
				public ServletContext getServletContext() {
					return JSPServlet.this.getServletContext();
				}
				public String getInitParameter(String name) {
					return null;
				}
				public Enumeration getInitParameterNames() {
					return (new Vector()).elements();
				}
			});
		} catch(Exception exc) {}

		jspServlets.put(className, srv);

		return srv;
	}

	private void unloadServlet(Servlet srv) {
		try {
			srv.destroy();
		} catch(Exception exc) { }

		jspServlets.remove(srv.getClass().getName());
	}

	private boolean checkServletDependencies(Class c) throws JSPException {
		Method			meth;

		try {
			// invoke dependency checking method
			meth = c.getMethod("__checkDependencies", new Class[0]);
			return ((Boolean) meth.invoke(null, new Object[0])).booleanValue();
		} catch(IllegalAccessException iaexc) {
			throw new JSPException("Jsp servlet class " + c.getName() + " does not have a public static __checkDependencies() method");
		} catch(InvocationTargetException itexc) {
			throw new JSPException("Jsp servlet class " + c.getName() + "'s __checkDependencies() method threw an exception: " + itexc.getTargetException());
		} catch(NoSuchMethodException nsfexc) {
			throw new JSPException("Jsp servlet class " + c.getName() + " does not have a __checkDependencies() method that returns a boolean");
		} catch(SecurityException secexc) {
			throw new JSPException("Jsp servlet class " + c.getName() + " has a security problem with its __checkDependencies() method");
		}
	}

	private void refreshClassLoader(Class c) {
		if(c.getClassLoader() == jspClassLoader) {
			jspClassLoader = new JSPClassLoader(classFilesDir);
		}
	}

	private static String urlDecode(String val) {
        StringBuffer	buf = new StringBuffer(val.length());
        char			c;

		for (int i = 0; i < val.length(); i++) {
			c = val.charAt(i);
			if(c == '%') {
				try {
					buf.append((char)Integer.parseInt(val.substring(i+1,i+3),16));
					i += 2;
					continue;
				} catch(Exception e) { }
			} else if(c == '+') {
				buf.append(' ');
				continue;
			}
 			buf.append(c);
		}
        return buf.toString();
	}
}
